from collections.abc import Sized, Iterator
import logging
import sys
from typing import Any, Callable, Optional, TypeVar, Generic, Iterable


def set_global(name: str, value: Any):
    globals()[name] = value


def get_global(name: str, default=None) -> Optional[Any]:
    try:
        return globals()[name]
    except KeyError:
        return None if default is None else default


def set_logger_name(name: str):
    globals()["_logger_name"] = name


def _get_logger_name() -> Optional[str]:
    try:
        logger_name = globals()["_logger_name"]
    except KeyError:
        logger_name = None
    return logger_name


def set_use_print(use_print: bool):
    globals()["_use_print"] = use_print


def _get_use_print() -> bool:
    try:
        use_print = globals()["_use_print"]
    except KeyError:
        use_print = False
    return use_print


def set_callback(callback: Callable):
    globals()["_callback"] = callback


def _get_callback() -> Optional[Callable]:
    try:
        callback = globals()["_callback"]
    except KeyError:
        callback = None
    return callback


def _get_logger() -> logging.Logger:
    logger_name = _get_logger_name()
    return logging.getLogger(logger_name)


def if_level(level: int, message: str):
    _use_print = _get_use_print()
    _callback = _get_callback()
    _logger = _get_logger()

    if _logger.level <= level:
        if _use_print:
            print(f"{message}", flush=True)
        if _callback:
            _callback(message)


def dbg(message: Any):
    message = str(message)
    logger = _get_logger()
    if_level(logging.DEBUG, message)
    logger.debug(message)


def log(message: Any):
    message = str(message)
    logger = _get_logger()
    if_level(logging.INFO, message)
    logger.info(message)


def info(message: Any):
    message = str(message)
    logger = _get_logger()
    if_level(logging.INFO, message)
    logger.info(message)


def warn(message: Any):
    message = str(message)
    logger = _get_logger()
    if_level(logging.WARNING, message)
    logger.warning(message)


def err(message: Any):
    message = str(message)
    logger = _get_logger()
    if_level(logging.ERROR, message)
    logger.error(message)


def basic_config():
    logging.basicConfig(
        format="%(asctime)s %(levelname)s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=logging.DEBUG,
        stream=sys.stdout,
    )


A = TypeVar("A")


class progress(Generic[A]):
    def __init__(
        self,
        iterator: Iterable[A],
        name: str = "",
        num_steps: int = 10,
        precision: int = 1,
        print_idx: bool = False,
    ):
        if not isinstance(iterator, Iterable):
            raise ValueError("Passed object is not iterable")
        if not isinstance(iterator, Sized):
            raise ValueError("Passed object does not have a size")

        self.print_idx = print_idx
        self.iterator: Iterator[A] = iter(iterator)
        self.len = len(iterator)
        self.name = name
        self.num_steps = num_steps
        if self.num_steps > self.len:
            # msg = "num_steps > len(iterator), setting num_steps to len(iterator)"
            # err(msg)
            self.num_steps = self.len

        self.map = {}
        step_size = 100 / self.num_steps
        is_int = step_size.is_integer()
        for i in range(self.num_steps + 1):
            index = i * int(self.len / self.num_steps)
            percent = i * step_size
            percent_str = f"{percent:.{precision}f}%"
            if is_int:
                percent_str = f"{int(percent):d}%"
            self.map[index] = percent_str

        self.indices = set(self.map.keys())
        self.index = 0
        self.index_len = len(str(self.len))
        self.percent_len = max([len(p) for p in self.map.values()])

    def __iter__(self):
        return self

    def __next__(self) -> A:
        if self.index in self.indices:
            msg = f"{self.name} {self.map[self.index]:>{self.percent_len}}"
            if self.print_idx:
                msg += f" n={self.index:<{self.index_len}}"
            log(msg)
        if self.index >= self.len:
            raise StopIteration
        else:
            self.index += 1
            return next(self.iterator)


if __name__ == "__main__":
    import time

    set_logger_name("progress")

    # setup basic logging format
    basic_config()

    # too lazy rn to make a proper test suite
    for i in progress(range(101), name="test1"):
        time.sleep(0.001)

    for i in progress(
        range(7),
        name="test2",
        print_idx=True,
        num_steps=11,
        precision=3,
    ):
        time.sleep(0.001)

    for idx, i in enumerate(progress(["a", "b", "c"], name="test3")):
        time.sleep(0.001)

    for i in progress([0.1, 0.2, 0.3], name="test4"):
        log(i)
        time.sleep(0.001)

    def rev(msg):
        print(msg[::-1])
        pass

    set_use_print(True)
    set_logger_name("ayy")
    set_callback(rev)
    log("hello world")

    logging.getLogger("ayy").setLevel(logging.INFO)
    dbg("debug message")

    set_global("ayy", "lmao")
    value = get_global("ayy")
    log(value)

    from datetime import datetime

    err(datetime.now())



import click
import subprocess


def get_list_of_python_versions():
    res = subprocess.run(
        ["pyenv", "install", "--list"],
        capture_output=True,
        text=True,
    )
    if res.returncode != 0:
        click.echo(res.stderr)
        return
    versions = res.stdout.splitlines()
    return [version.strip() for version in versions[1:]]


@click.group(name="utilki")
def cli():
    pass


@cli.command()
@click.argument(
    "python_version",
    type=click.Choice(get_list_of_python_versions()),
    default="3.8.10",
)
def venv(python_version):
    # Create the virtual environment with the given name and Python version
    venv_name = click.prompt("Enter venv name", type=str)
    res = subprocess.run(
        ["pyenv", "virtualenv", python_version, venv_name],
        capture_output=True,
        text=True,
    )
    if res.returncode != 0:
        if "is not installed" in res.stderr:
            response = click.prompt(
                f"Python version {python_version} is not installed. Install it?",
                type=click.Choice(["y", "n"]),
                default="y",
            )
            if response == "y":
                process = subprocess.Popen(
                    ["pyenv", "install", python_version],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    text=True,
                )
                for line in iter(process.stdout.readline, ""):  # type: ignore
                    click.echo(line.strip())
                return_code = process.wait()
                if return_code != 0:
                    click.echo("Installation failed")
                    return
                click.echo("Installation successful")

                res = subprocess.run(
                    ["pyenv", "virtualenv", python_version, venv_name],
                    capture_output=True,
                    text=True,
                )
                if res.returncode != 0:
                    click.echo(res.stderr)
                    return

    click.echo(
        f"Created venv `{venv_name}` with Python version {python_version}"
    )


if __name__ == "__main__":
    cli()



import json
import os
from datetime import datetime
from dataclasses import Field
from pydantic.fields import ModelField
from typing import (
    Any,
    ClassVar,
    Dict,
    Iterable,
    List,
    Tuple,
    Union,
)


Defaults = Union[
    datetime,
    bool,
    int,
    float,
    str,
    List[bool],
    List[int],
    List[float],
    List[str],
]


singles = [int, float, str, bool, datetime]
lists = [List[int], List[float], List[str], List[bool]]
options = [
    Union[int, None],
    Union[float, None],
    Union[str, None],
    Union[bool, None],
]

dicts = [Dict[str, Any], Dict[str, str], Dict[str, int], Dict[str, float]] + [
    Dict[str, Dict[str, Any]]
]

types_we_support = singles + lists + options + dicts

IsDefault = bool


class TaskMixin:
    __dataclass_fields__: ClassVar[Dict[str, Any]]
    __fields__: ClassVar[Dict[str, Any]]

    @classmethod
    def __init__(cls, **kwargs):
        ...

    @classmethod
    def create(cls):
        params: List[Tuple[str, Any]] = []
        if hasattr(cls, "__dataclass_fields__"):
            fields: Iterable[Field] = cls.__dataclass_fields__.values()
            for field in fields:
                params.append((field.name, field.type))
        elif hasattr(cls, "__fields__"):
            model_fields: Iterable[ModelField] = cls.__fields__.values()
            for model_field in model_fields:
                params.append((model_field.name, model_field.annotation))
        else:
            raise TypeError("Invalid type")
        task = cls(**{name: cls.parse(name, type) for name, type in params})
        return task

    @classmethod
    def get_default(cls, name_) -> Tuple[IsDefault, Defaults]:
        env_var = os.getenv(name_)
        if env_var is not None:
            return False, env_var
        if hasattr(cls, "__dataclass_fields__"):
            field: Field = cls.__dataclass_fields__[name_]
            result: Any = field.default
            return True, result
        elif hasattr(cls, "__fields__"):
            model_field: ModelField = cls.__fields__[name_]
            default_model = model_field.get_default()
            return True, default_model
        else:
            raise TypeError("Invalid type")

    @classmethod
    def parse(cls, name_, type_):
        is_default, value = cls.get_default(name_)
        if is_default:
            if type_ not in types_we_support:
                raise TypeError("Invalid type")
            if type_ != type(value) and type_ in singles:
                raise TypeError("Invalid type")
            if type_ in lists and not isinstance(value, list):
                raise TypeError("Invalid type")
            else:
                return value
        if isinstance(value, str):
            res = parse_variations(type_, value, name_)
            return res
        elif isinstance(value, list):
            if type_ in [List[int], List[str], List[float], List[bool]]:
                return value
        elif isinstance(value, datetime):
            if type_ == datetime:
                return value
        elif type_ in [int, str, float, bool]:
            if isinstance(value, type_):
                return value
        else:
            raise TypeError("Invalid type")


def parse_bool(param):
    if param in ["True", "true", True]:
        return True
    elif param in ["False", "false", False]:
        return False
    else:
        raise TypeError("Invalid boolean format")


def parse_options(value, type_):
    if value in ["None", "none", None, "null", "NULL", ""]:
        return None
    else:
        return type_(value)


def parse_list(list_str: str, type_) -> List[Any]:
    if list_str.startswith("[") and list_str.endswith("]"):
        return json.loads(list_str)
    # TODO: make this a separate tuple parsing function
    # elif list_str.startswith("(") and list_str.endswith(")"):
    #     list_str.replace("(", "[")
    #     list_str.replace(")", "]")
    #     return json.loads(list_str)
    elif "," in list_str:
        return list(map(type_, list_str.split(",")))
    else:
        raise ValueError("Invalid list format")


def get_date(param: str):
    n = [int(n) for n in param.split("-")]
    if len(n) == 3:
        return datetime(n[0], n[1], n[2])
    elif len(n) == 6:
        return datetime(n[0], n[1], n[2], n[3], n[4], n[5])
    else:
        raise ValueError("Invalid datetime format")


def parse_variations(type_, value, name_):
    if type_ == List[int]:
        return parse_list(value, int)
    elif type_ == List[str]:
        return parse_list(value, str)
    elif type_ == List[float]:
        return parse_list(value, float)
    elif type_ == List[bool]:
        return parse_list(value, parse_bool)
    elif type_ == Union[int, None]:
        return parse_options(value, int)
    elif type_ == Union[str, None]:
        return parse_options(value, str)
    elif type_ == Union[float, None]:
        return parse_options(value, float)
    elif type_ == Union[bool, None]:
        return parse_options(value, parse_bool)
    elif type_ == Dict[int, Any]:
        return json.loads(value)
    elif type_ == Dict[str, Any]:
        return json.loads(value)
    elif type_ == Dict[float, Any]:
        return json.loads(value)
    elif type_ == Dict[bool, Any]:
        return json.loads(value)
    elif type_ == Dict[str, Dict[str, Any]]:
        return json.loads(value)
    elif type_ == bool:
        return parse_bool(value)
    elif type_ == int:
        return int(value)
    elif type_ == float:
        return float(value)
    elif type_ == str:
        try:
            res = json.loads(value)
            if type(res) == int:
                return str(res)
            else:
                return res
        except Exception:
            return str(value)
    elif type_ == datetime:
        # FIXME actually like use proper parsing lmao
        if value.startswith('"') and value.endswith('"'):
            value = value[1:-1]
        has_t_or_colon = ":" in value or "T" in value
        num_parts = len(value.split("-"))
        if has_t_or_colon:
            value = value.replace(" ", "-")
            value = value.replace("T", "-")
            value = value.replace(":", "-")
            return get_date(value)
        elif num_parts == 3:
            return get_date(value)
        else:
            raise TypeError("Invalid datetime format")



# flake8: noqa

__version__ = "0.1.0"

from .task_mixin import TaskMixin
from .log_utils import (
    progress,
    set_logger_name,
    set_use_print,
    set_callback,
    log,
    dbg,
    info,
    err,
    warn,
    basic_config,
    set_global,
    get_global,
)
